#define FIG_INT_CNT		4

#define fig_USA			0
#define fig_turn		1
#define fig_set			2
#define fig_turnSingle		3
#define fig_turnSingleCloverleaf	4
#define fig_gypsy		5
#define fig_circle		6
#define fig_cast		7
#define fig_lead		8
#define fig_figureEight		9
#define fig_doubleFigureEight	10
#define fig_handsAcross		11
#define fig_openLadiesChain	12
#define fig_poussette		13
#define fig_drawPoussette	14
#define fig_chevron		15
#define fig_hey			16
#define fig_mirrorHey		17
#define fig_parallelHey		18
#define fig_interwovenHey	19
#define fig_progressiveHey	20
#define fig_heyFor3		21
#define fig_heyFor4		22
#define fig_dolphinHey		23
#define fig_circularHey		24
#define fig_rightsAndLefts	25
#define fig_tandemHey		26
#define fig_castAndLead		27
#define fig_leadAndCast		28
#define fig_madRobin		29
#define fig_californiaTwirl	30
#define fig_boxTheGnat		31
#define fig_swatTheFlea		32
#define fig_grandSquare		33
#define fig_setAndLink		34
#define fig_bentleyAllemande	35
#define fig_scottishAllemande	36
#define fig_scottishPromenade	37
#define fig_scottishKnot	38
#define fig_diamondPoussette	39
#define fig_setAndRotate	40
#define fig_squareLadiesChain	41
#define fig_rightAndLeftThrough	42
#define fig_bendTheLine		43
#define fig_oceanWave		44
#define fig_swingThrough	45
#define fig_petronella		46
#define fig_starPromenade	47
#define fig_butterflyWhirl	48
#define fig_longLinesForwardAndBack	49
#define fig_kiss		50
#define fig_regencyAllemande	51
#define fig_leadOutSides	52
#define fig_swingCorners	53
#define fig_swing		54
#define fig_waltz		55
#define fig_siding		56
#define fig_arming		57
#define fig_rightAndLeft4	58
#define fig_polka		59
#define fig_crossGoBelow	60
#define fig_downTheOutsideBack	61
#define fig_downTheMiddleBack	62
#define fig_rondel		63
#define fig_doubleTriangle	64
#define fig_espagnol		65
#define fig_helloGoodbye	66
#define fig_setAndTurn		67
#define fig_snakePass		68
#define fig_schiehallionReel	69
#define fig_courageReel		70
#define fig_contraCorners	71
#define fig_back2back		72
#define fig_face2face		73
#define fig_chevronPoussette	74


extern void figureMask(Dance *dance,unsigned int masks[FIG_INT_CNT]);
extern int danceHasStarBack(Dance *dance);
