/* splashwin.c */
extern void on_Splash_New_clicked(GtkWidget *button, gpointer ud);
extern void on_Splash_Open_clicked(GtkWidget *button, gpointer ud);
extern void on_Splash_Quit_clicked(GtkWidget *button, gpointer ud);
extern void on_Splash_Calls_clicked(GtkWidget *button, gpointer ud);
extern void on_Splash_destroy(GtkWidget *info, gpointer ud);
/* opennewdlg.c */
extern void on_NewFigure_Starting_changed(GtkWidget *combobox, gpointer ud);
extern void on_NewFigure_Style_changed(GtkWidget *combobox, gpointer ud);
extern void on_NewDanceFigure_toggled(GtkWidget *dlg, gpointer ud);
extern void on_NewDlg_realize(GtkWidget *combobox, gpointer ud);
extern void on_PickAFigure_rowActivated(GtkWidget *tv, gpointer ud);
